@@ -283,7 +283,7 @@ ftp_response_to_vfs_result (FtpConnection *conn)
 }
 
 #ifdef HAVE_GSSAPI
-static char *radixN =
+static const char *radixN =
 	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
 static char pad = '=';
@@ -1011,9 +1011,7 @@ ftp_kerberos_login (FtpConnection *conn,
 	char *encoded_token;
 	int len;
 	GnomeVFSResult result;
-	gboolean auth_ok;
 
-	auth_ok = FALSE;
 	result = do_basic_command (conn, "AUTH GSSAPI", cancellation);
 	if (result != GNOME_VFS_OK) {
 		return result;
@@ -1868,7 +1866,52 @@ do_create (GnomeVFSMethod *method,
 	   guint perm,
 	   GnomeVFSContext *context)
 {
-	return do_open(method, method_handle, uri, mode, context);
+	GnomeVFSResult result;
+	FtpConnection *conn;
+	gchar *chmod_command;
+
+	if ((mode & GNOME_VFS_OPEN_READ) && (mode & GNOME_VFS_OPEN_WRITE)) {
+		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
+	}
+
+	if (!(mode & GNOME_VFS_OPEN_READ) && !(mode & GNOME_VFS_OPEN_WRITE)) {
+		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
+	}
+
+	result = ftp_connection_acquire (uri, &conn, context);
+	if (result != GNOME_VFS_OK)
+		return result;
+
+	/* Exclusive set: we have to fail if the file exists */
+	if (exclusive) {
+		conn->operation = FTP_READ;
+		result = do_path_transfer_command (conn, "RETR", uri, context);
+
+		/* File exists or an error occurred */
+		if (result != GNOME_VFS_ERROR_NOT_FOUND) {
+			ftp_connection_release (conn, TRUE);
+
+			/* Return the error if there is one, otherwise the file exists */
+			return (result == GNOME_VFS_OK) ? GNOME_VFS_ERROR_FILE_EXISTS
+		                                        : result;
+		}
+	}
+
+	result = do_open(method, method_handle, uri, mode, context);
+	if (result != GNOME_VFS_OK) {
+		ftp_connection_release (conn, TRUE);
+		return result;
+	}
+
+	/* This is a not-standard command and FTP server may not support it,
+	 * so eventual errors will be ignored */
+	chmod_command = g_strdup_printf("SITE CHMOD %o", perm);
+	do_path_command (conn, chmod_command, uri, get_cancellation (context));
+
+	g_free (chmod_command);
+	ftp_connection_release (conn, TRUE);
+
+	return result;
 }
 
 static GnomeVFSResult 
@@ -2192,10 +2235,14 @@ unix_ls_to_file_info (gchar *ls, GnomeVFSFileInfo *file_info,
 
 		if (file_info->type == GNOME_VFS_FILE_TYPE_REGULAR) {
 			mime_type = gnome_vfs_mime_type_from_name_or_default (file_info->name, GNOME_VFS_MIME_TYPE_UNKNOWN);
-			/*ftp_debug (conn, g_strdup_printf ("mimetype = %s", mime_type));*/
 		} else {
-			mime_type = gnome_vfs_mime_type_from_mode (s.st_mode);
+			/* TODO replace this by gnome_vfs_mime_type_from_mode_or_default call, #330625 */
+			mime_type = gnome_vfs_mime_type_from_mode (file_info->permissions);
+			if (mime_type == NULL) {
+				mime_type = GNOME_VFS_MIME_TYPE_UNKNOWN;
+			}
 		}
+		/*ftp_debug (conn, g_strdup_printf ("mimetype = %s", mime_type));*/
 		file_info->mime_type = g_strdup (mime_type);
 		file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
 
@@ -2349,6 +2396,38 @@ do_get_file_info (GnomeVFSMethod *method,
 		}
 		g_free (name);
 		do_close_directory (method, method_handle, context);
+
+		/* maybe it was a hidden directory not included in ls output
+		 * (IIS virtual directory), so try to CWD to it. */
+		if (result == GNOME_VFS_ERROR_NOT_FOUND) {
+			FtpConnection *connection;
+
+			if (ftp_connection_acquire (uri, &connection, context) == GNOME_VFS_OK) {
+				result = do_path_command (connection, "CWD", uri, get_cancellation (context));
+				ftp_connection_release (connection, FALSE);
+			}
+
+			if (result == GNOME_VFS_OK) {
+				char *unescaped;
+				char *basename;
+
+				unescaped = gnome_vfs_unescape_string (uri->text, G_DIR_SEPARATOR_S);
+				basename = g_path_get_basename (unescaped);
+				g_free (unescaped);
+
+				if (basename != NULL) {
+					file_info->name = basename;
+					file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
+					file_info->mime_type = g_strdup ("x-directory/normal");
+					file_info->valid_fields = GNOME_VFS_FILE_INFO_FIELDS_TYPE |
+						GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
+				} else {
+					result = GNOME_VFS_ERROR_NOT_FOUND;
+				}
+			} else {
+				result = GNOME_VFS_ERROR_NOT_FOUND;
+			}
+		}
 	}
 
 	return result;
@@ -2604,7 +2683,7 @@ do_read_directory (GnomeVFSMethod *method,
 					break;
 				}
 				
-				symlink_name = g_strdup (symlink_info->symlink_name);
+				symlink_name = gnome_vfs_escape_path_string (symlink_info->symlink_name);
 				gnome_vfs_file_info_clear (symlink_info);
 				
 				link_uri = gnome_vfs_uri_resolve_relative (uri, symlink_name);
@@ -2695,6 +2774,15 @@ do_make_directory (GnomeVFSMethod *method,
 	GnomeVFSResult result;
 	gchar *chmod_command;
 
+	result = do_path_command_completely ("CWD", uri, context,
+			/* If the directory doesn't exist @result will be !GNOME_VFS_OK,
+			* in this case we don't return and proceed with MKD */
+			!GNOME_VFS_OK);
+
+	if (result == GNOME_VFS_OK) {
+		return GNOME_VFS_ERROR_FILE_EXISTS;
+	}
+
 	result = do_path_command_completely ("MKD", uri, context, 
 		GNOME_VFS_ERROR_ACCESS_DENIED);
 
